#!/usr/bin/env python3
# -*- coding: utf-8; -*-




import katoptron as m
import tbox
import tbox.gmsh as gmsh
import katoptron.utilities as u
from PyTrilinos import Teuchos
import shlex
import subprocess
import os
import numpy as np
import tarfile
import glob
from fwk.wutils import parseargs
from fwk.testing import *
from katoptron.readers import read_Belos


def model(msh,
          E,
          p,
          ensemble_size):
    comm, rank, size = m.utilities.import_MPI()

    pbl = m.Problem(msh, comm)

    nu = 0.2 * np.ones(ensemble_size)
    k = 0.114
    d = 0.

    zero = np.zeros(ensemble_size)
    m.Medium(pbl, "Body 1", "test", E, nu, k, d)

    m.Dirichlet(pbl, "Surf 1 1", "Clamped", 1, 0., 1, 0., 1, 0., 1, 0.,
                ensemble_size)
    m.Neumann(pbl, "Surf 1 2", "heat fluxes", 0, zero, 0, zero, 1, p, 0, zero,
              ensemble_size)

    solverList = Teuchos.ParameterList()
    solverList['Convergence Tolerance'] = 10**(-8)
    solverList['Maximum Iterations'] = 100
    solverList['Verbosity'] = 33
    solverList['Flexible Gmres'] = False
    solverList['Output Style'] = 1
    solverList['Output Frequency'] = 1

    solverList['Use preconditioner'] = False

    solverList['type'] = "BlockGmres"
    solverList['Maximum active set iteration'] = 1

    solverList['Print Teuchos timers'] = True
    solverList['Write txt files'] = True
    solverList['Write matrix and vectors'] = True

    slv = m.IterativeSolver(pbl, solverList, 3, ensemble_size)

    slv.start()


def main():
    comm, rank, size = u.import_MPI()

    useER = m.UseEnsembleReduction()

    geo_name = '1_cube.geo'
    file_dir = os.path.dirname(__file__)
    work_dir = os.getcwd()

    msh, mesh_name, partitioned_mesh_name = u.read_mesh_MPI(
        geo_name, file_dir, work_dir, comm, rank, size)

    E_mean = 1.
    E_min = 0.75
    E_max = 1.25

    E_min_2 = 0.95
    E_max_2 = 1.05

    E_min_3 = 1.2
    E_max_3 = 1.3

    p_mean = 0.1
    p_min = 0.075
    p_max = 0.125

    ensemble_size = 8
    E = np.linspace(E_min, E_max, ensemble_size)
    p = p_mean * np.ones(ensemble_size)

    nThreads = u.Initialize_Kokkos()
    model(msh, E, p, ensemble_size)

    u.Finalize_Kokkos()

    if rank == 0:
        if useER:
            iter_indices, residuals = read_Belos(work_dir+'/belos_out.txt', 1)

            tests = CTests()
            nIterations = iter_indices[-1]
            nExpectedIterations = 75
            tests.add(CTest('Number of iterations with ensemble reduction',
                            nIterations, nExpectedIterations, 0.))
            tests.run()
        else:
            iter_indices, residuals = read_Belos(
                work_dir+'/belos_out.txt', ensemble_size)

            tests = CTests()
            nIterations = iter_indices[-1]
            nExpectedIterations = 35
            tests.add(CTest('Number of iterations without ensemble reduction',
                            nIterations, nExpectedIterations, 0.))
            tests.run()


if __name__ == "__main__":
    main()
