/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wElement.h"
#include "wNode.h"
#include "wTag.h"
#include "wCache.h"
#include "std_extra.h"

using namespace tbox;

namespace tbox
{

TBOX_API std::ostream &
operator<<(std::ostream &out, ELTYPE const &obj)
{
    switch (obj)
    {
    case ELTYPE::LINE2:
        out << "LINE2";
        break;
    case ELTYPE::TRI3:
        out << "TRI3";
        break;
    case ELTYPE::QUAD4:
        out << "QUAD4";
        break;
    case ELTYPE::TETRA4:
        out << "TETRA4";
        break;
    case ELTYPE::HEX8:
        out << "HEX8";
        break;
    case ELTYPE::POINT1:
        out << "POINT1";
        break;
    default:
        out << "UNKNOWN-" << static_cast<int>(obj);
        break;
    }
    return out;
}

} // namespace tbox

Element::Element(int n, Tag *_ptag, Tag *_etag,
                 std::vector<Tag *> &_parts, std::vector<Node *> &nods)
    : wObject(), order(0), hasVals(false), hasGrads(false), isVol(false),
      no(n), ptag(_ptag), etag(_etag), parts(_parts), nodes(nods)
{
    if (ptag)
        ptag->elems.push_back(this);
    if (etag)
        etag->elems.push_back(this);

    for (auto p : parts)
        p->elems.push_back(this);
}

void Element::initValues(bool isvol, int ordr)
{
    throw std::runtime_error("Element::initValues() not implemented!\n");
}
void Element::initGradients()
{
    throw std::runtime_error("Element::initGradients() not implemented!\n");
}
void Element::update()
{
    throw std::runtime_error("Element::update() not implemented!\n");
}

/**
 * @brief Check that determinant of Jacobian is positive
 */
void Element::checkJac(size_t k) const
{
    if (detJs[k] <= 0)
    {
        std::stringstream err;
        err << "Negative jacobian detected for element: " << *this << "\n";
        throw std::runtime_error(err.str());
    }
}

void Element::buildJ()
{
    throw std::runtime_error("Element::buildJ not implemented\n");
}
void Element::buildSurfaceJ()
{
    throw std::runtime_error("Element::buildSurfaceJ not implemented\n");
}
void Element::buildGradientJ()
{
    throw std::runtime_error("Element::buildGradientJ not implemented\n");
}
void Element::buildGradientSurfaceJ()
{
    throw std::runtime_error("Element::buildGradientSurfaceJ not implemented\n");
}

void Element::computeV()
{
    throw std::runtime_error("Element::computeV not implemented\n");
}
void Element::buildGradientV(int dim)
{
    throw std::runtime_error("Element::buildGradientV not implemented\n");
}

/**
 * @brief Compute center of gravity of element
 */
void Element::computeCg()
{
    cg = Eigen::Vector3d::Zero();
    for (auto n : nodes)
        cg += n->pos;
    cg /= static_cast<double>(nodes.size());
}

std::vector<Eigen::Vector3d> Element::computeTangents() const
{
    throw std::runtime_error("Element::computeTangents not implemented\n");
}
std::vector<Eigen::Vector3d> Element::buildTangents(size_t k) const
{
    throw std::runtime_error("Element::buildTangents not implemented\n");
}

void Element::computeNormal()
{
    throw std::runtime_error("Element::computeNormal not implemented\n");
}
void Element::computeGradientNormal()
{
    throw std::runtime_error("Element::computeGradientNormal not implemented\n");
}

/**
 * @brief Return the values extrapolated at all the element nodes
 * based on the values at Gauss points
 */
Eigen::MatrixXd Element::gp2node(Eigen::MatrixXd &values_at_gp) const
{
    Cache &cache = getVCache();
    size_t ngp = cache.getVGauss().getN();

    Eigen::MatrixXd A(ngp, nodes.size());
    for (size_t i = 0; i < ngp; ++i)
        A.row(i) = cache.getSf(i).transpose();

    return (A.transpose() * A).inverse() * A.transpose() * values_at_gp; // nodes.size() * 18
}

Eigen::VectorXd Element::computeGradient(std::vector<double> const &u, size_t k) const
{
    throw std::runtime_error("Element::computeGradient not implemented\n");
}

void Element::write(std::ostream &out) const
{
    out << type() << " #" << no << ":";

    out << " nodes=[ ";
    for (auto n : nodes)
        out << "#" << n->no << ' ';
    out << "]";

    if (ptag)
        out << " ptag #" << ptag->no;
    if (etag)
        out << " etag #" << etag->no;
    if (!parts.empty())
    {
        out << " parts=[ ";
        for (auto p : parts)
            out << "#" << p->no << ' ';
        out << "]";
    }
}
